/*
**	Apple Macintosh Developer Technical Support
**
**	A fairly flexible, international friendly Standard File "Get Folder"
**	dialog that handles all the goofy things I could find to work-around.
**
**	by Jim Luther, Apple Developer Technical Support
**	based somewhat on code by Greg Robbins and Steve Falkenburg
**
**	File:		GetFolder.c
**
**	Copyright  1992-1994 Apple Computer, Inc.
**	All rights reserved.
**
**	You may incorporate this sample code into your applications without
**	restriction, though the sample code has been provided "AS IS" and the
**	responsibility for its operation is 100% yours.  However, what you are
**	not permitted to do is to redistribute the source as "DSC Sample Code"
**	after having made changes. If you're going to re-distribute the source,
**	we require that you make it clear in the source that the code was
**	descended from Apple Sample Code, but that you've made changes.
*/
/*
	Change History (most recent first):
		
		<2>		 8/29/98	Neb		Fixed no-response bug with File Sharing.
		<1>		--/--/--	JL		Original version from Jim Luther.
*/

#include <assert.h>
#include <LowMem.h>
#include <Gestalt.h>
#include <Memory.h>
#include <TextUtils.h>
#include <Folders.h>
#include <Aliases.h>
#include <Finder.h>
#include <StandardFile.h>
#include "GetFolder.r.h"
#include "GetFolder.h"

/*****************************************************************************/

/*
**	Local Constants
*/

enum
{
	kSelectFolderItem = 10,		/* select folder button item number */
	kSelectDesktopItem = 11,	/* select desktop button item number */
	kPromptItem = 12,			/* prompt item number */

	kBeforeFolderStrNum = 1,	/* text before directory name in folder select button */
	kAfterFolderStrNum = 2,		/* text after directory name in folder select button */
	kBeforeDesktopStrNum = 3,	/* text before directory name in desktop select button */
	kAfterDesktopStrNum = 4,	/* text after directory name in desktop select button */
	kDesktopStrNum = 5,			/* word 'Desktop' for folder select button */
	
	vcbAtrbLockMask = 0x8080,	/* bit 15 in ioVAtrb = if set, volume is software locked; */
								/* bit 7 in ioVAtrb = if set, volume is hardware locked */
	acUserMakeChanges = 0x04	/* bit 2 in ioACUser = if set, user does not */
								/*		have "Make Changes" privileges */
};

/*****************************************************************************/

/*
**	The SFData record contains the "global" data that the dialog hook function,
**	the file filter function, or the modal-dialog filter need to access.
*/

struct SFData
{
	StandardFileReply *SFRPtr;		/* Pointer to reply record */
	FSSpec	oldChoice;				/* The selection seen last by the dialog hook */
	Boolean	showInvisibles;			/* If TRUE, show invisible folders */
	Boolean	selectOnlyChangable;	/* If TRUE, select only changable (writable) folders */
	Str255	beforeFolderStr;		/* The string before the folder name in the folder select button */
	Str255	afterFolderStr;			/* The string after the folder name in the folder select button */
	Str255	beforeDesktopStr;		/* The string before the volume name in the desktop select button */
	Str255	afterDesktopStr;		/* The string after the volume name in the desktop select button */
	Str255	desktopStr;				/* The name to use for the desktop folder in the folder select button */
};
typedef struct SFData SFData;
typedef SFData *SFDataPtr;

/*****************************************************************************/

/*
**	Local Prototypes
*/

static	void	Concat(StringPtr dest,
					   StringPtr src);

static	void	SetButtonName(DialogPtr theDialog,
							  short buttonID,
							  ConstStr255Param choiceName,
							  SFDataPtr myDataPtr);

static	void	ActivateButton (DialogPtr theDlgPtr,
								short buttonID);

static	void	InactivateButton (DialogPtr theDlgPtr,
								  short buttonID);

static	Boolean	SameFSSpec(const FSSpec *itemFSSpec1,
						   const FSSpec *itemFSSpec2);

static	pascal	Boolean	MyCustomFileFilter(CInfoPBPtr myCInfoPBPtr,
										   SFDataPtr myDataPtr);

static	OSErr	SetSelectFolderButton(DialogPtr theDialog,
									  SFDataPtr myDataPtr);

static	OSErr	SetSelectDesktopButton(DialogPtr theDialog,
									   SFDataPtr myDataPtr);

static	OSErr	LastCallCleanup(SFDataPtr myDataPtr);

static	pascal	short	MyDlgHook(short item,
								  DialogPtr theDialog,
								  SFDataPtr myDataPtr);

/*****************************************************************************/

/*
**	Concat catenates two pascal-strings (Str255).
*/

static	void	Concat(StringPtr dest,
					   StringPtr src)
{
	short	copyLength;
	
	copyLength = src[0];
	if ( (copyLength + dest[0]) > 255 )
		copyLength = 255 - dest[0];
	
	if ( copyLength != 0 )
	{
		BlockMoveData(&src[1], &dest[dest[0] + 1], copyLength);
		dest[0] += copyLength;
	}
}

/*****************************************************************************/

/*
**	SetButtonName changes the name of the folder select button or 
**	the desktop select button to reflect the current selection.
*/

static	void	SetButtonName(DialogPtr theDialog,
							  short buttonID,
							  ConstStr255Param choiceName,
							  SFDataPtr myDataPtr)
{
	short	itemType;
	Handle	item;
	Rect	box;
	short	textWidth;
	OSErr	retCode;
	Str255	tempStr, tempStr2;
	
	BlockMoveData(choiceName, tempStr, choiceName[0] + 1);
	GetDialogItem(theDialog, buttonID, &itemType, &item, &box);
	switch ( buttonID )
	{
	case kSelectFolderItem:
		textWidth = (box.right - box.left - 2) -
					(StringWidth(myDataPtr->beforeFolderStr) +
					StringWidth(myDataPtr->afterFolderStr));
		retCode = TruncString(textWidth, tempStr, smTruncMiddle);
		tempStr2[0] = 0;
		Concat(tempStr2, myDataPtr->beforeFolderStr);
		Concat(tempStr2, tempStr);
		Concat(tempStr2, myDataPtr->afterFolderStr);
		SetControlTitle((ControlHandle)item, tempStr2);
		break;
	case kSelectDesktopItem:
		textWidth = (box.right - box.left - 2) -
					(StringWidth(myDataPtr->beforeDesktopStr) +
					StringWidth(myDataPtr->afterDesktopStr));
		retCode = TruncString(textWidth, tempStr, smTruncMiddle);
		tempStr2[0] = 0;
		Concat(tempStr2, myDataPtr->beforeDesktopStr);
		Concat(tempStr2, tempStr);
		Concat(tempStr2, myDataPtr->afterDesktopStr);
		SetControlTitle((ControlHandle)item, tempStr2);
		break;
	}
	ValidRect(&box); /* avoids flickering due to unnecessary redrawing */
}

/*****************************************************************************/

/*
**	ActivateButton makes the button active.
*/

static	void	ActivateButton(DialogPtr theDlgPtr,
							   short buttonID)
{
	short	itemType;
	Handle	item;
	Rect	box;

	GetDialogItem(theDlgPtr, buttonID, &itemType, &item, &box);
	HiliteControl((ControlHandle)item, 0);
}

/*****************************************************************************/

/*
**	InactivateButton makes the button inactive.
*/

static	void	InactivateButton(DialogPtr theDlgPtr,
								 short buttonID)
{
	short	itemType;
	Handle	item;
	Rect	box;

	GetDialogItem(theDlgPtr, buttonID, &itemType, &item, &box);
	HiliteControl((ControlHandle)item, 255);
}

/*****************************************************************************/

/*
**	SameFSSpec checks to FSSpec records to see if they are the same.
*/

static	Boolean	SameFSSpec(const FSSpec *itemFSSpec1,
						   const FSSpec *itemFSSpec2)
{
	return ( (itemFSSpec1->vRefNum == itemFSSpec2->vRefNum) && 
			 (itemFSSpec1->parID == itemFSSpec2->parID) && 
	       	 (EqualString(itemFSSpec1->name,itemFSSpec2->name, false, true)) );
}

/*****************************************************************************/

/*
**	MyCustomFileFilter filters out everything but directories from the list.
**	It also filters invisible directories if requested.
*/

static	pascal	Boolean	MyCustomFileFilter(CInfoPBPtr myCInfoPBPtr,
										   SFDataPtr myDataPtr)
{
	Boolean result = true;
	
	/* is this a directory? */
	if ( (myCInfoPBPtr->hFileInfo.ioFlAttrib & ioDirMask) != 0 )
	{
		if ( myDataPtr->showInvisibles )
			result = false;	/* show invisibles, too */
		else
			/* filter out invisibles */
			result = ((myCInfoPBPtr->dirInfo.ioDrUsrWds.frFlags & kIsInvisible) != 0);
	}
	return result;
}

/*****************************************************************************/

/*
**	SetSelectFolderButton activates or deactivates the Select "xxx" button
*/

static	OSErr	SetSelectFolderButton(DialogPtr theDialog,
									  SFDataPtr myDataPtr)
{
	Str63			folderName;
	short			desktopVRefNum;
	long			desktopDirID;
	HParamBlockRec	hPB;
	CInfoPBRec		cPB;
	OSErr			retCode;

	BlockMoveData(myDataPtr->SFRPtr->sfFile.name,
				  folderName,
				  myDataPtr->SFRPtr->sfFile.name[0] + 1);
	
	/* get info on current choice */
	cPB.dirInfo.ioVRefNum = myDataPtr->SFRPtr->sfFile.vRefNum;
	cPB.dirInfo.ioDrDirID = myDataPtr->SFRPtr->sfFile.parID;
	cPB.dirInfo.ioNamePtr = (StringPtr)folderName;
	if ( folderName[0] != 0 )
		cPB.dirInfo.ioFDirIndex = 0;	/* get info on named directory */
	else
		cPB.dirInfo.ioFDirIndex = -1;	/* get info on dirID and return folderName */
	cPB.dirInfo.ioACUser = 0;	/* Note: ioACUser is filler2 in old headers */
	retCode = PBGetCatInfoSync(&cPB);
	if ( retCode == noErr )
	{
		if ( FindFolder(myDataPtr->SFRPtr->sfFile.vRefNum, kDesktopFolderType,
						kCreateFolder, &desktopVRefNum, &desktopDirID) != noErr )
		{
			/* if desktop folder isn't found, set desktopDirID to value that */
			/* won't match any parID */
			desktopDirID = 0;	
		}
		if ( (myDataPtr->SFRPtr->sfFile.parID == desktopDirID) &&
			 (myDataPtr->SFRPtr->sfFile.name[0] == 0) )
		{
			/* At desktop level with no items selected, */
			/* use desktopStr as folder name */
			BlockMoveData(myDataPtr->desktopStr, folderName, myDataPtr->desktopStr[0] + 1);
		}
		
		SetButtonName(theDialog, kSelectFolderItem, folderName, myDataPtr);
		
		/* Should choose folder button be active? */
		hPB.volumeParam.ioNamePtr = NULL;
		hPB.volumeParam.ioVRefNum = myDataPtr->SFRPtr->sfFile.vRefNum;
		hPB.volumeParam.ioVolIndex = 0;	/* use vRefNum only */
		retCode = PBHGetVInfoSync(&hPB);

		if ( retCode == noErr )
		{
			if ( !myDataPtr->selectOnlyChangable ||
			     ((hPB.volumeParam.ioVAtrb & vcbAtrbLockMask) == 0) &&
			      ((cPB.dirInfo.ioACUser & acUserMakeChanges) == 0) )	/* Note: ioACUser is filler2 in old headers */
			{
				ActivateButton(theDialog, kSelectFolderItem);
			}
			else
			{
				InactivateButton(theDialog, kSelectFolderItem);
			}
		}
	}
	return retCode;
}

/*****************************************************************************/

/*
**	SetSelectDesktopButton activates or deactivates the Select Desktop on "xxx"
**	button.
*/

static	OSErr	SetSelectDesktopButton(DialogPtr theDialog,
									   SFDataPtr myDataPtr)
{
	Str63			volName;
	short			desktopVRefNum;
	long			desktopDirID;
	HParamBlockRec	hPB;
	CInfoPBRec		cPB;
	OSErr			retCode;
	Boolean			hasDesktopFolder;
	
	/* If volume has changed, set choose desktop button name */
	retCode = noErr;
	if ( myDataPtr->SFRPtr->sfFile.vRefNum != myDataPtr->oldChoice.vRefNum )
	{
		hPB.volumeParam.ioNamePtr = volName;
		hPB.volumeParam.ioVRefNum = myDataPtr->SFRPtr->sfFile.vRefNum;
		hPB.volumeParam.ioVolIndex = 0;	/* use vRefNum only */
		retCode = PBHGetVInfoSync(&hPB);
		if ( retCode == noErr )
		{
			SetButtonName(theDialog, kSelectDesktopItem, volName, myDataPtr);
		}
	}
	
	if ( retCode == noErr )
	{
		hasDesktopFolder = ( FindFolder(myDataPtr->SFRPtr->sfFile.vRefNum,
										kDesktopFolderType, kCreateFolder, &desktopVRefNum,
										&desktopDirID) == noErr );
		if ( hasDesktopFolder )
		{
			/* get info on current desktop choice */
			cPB.dirInfo.ioVRefNum = desktopVRefNum;
			cPB.dirInfo.ioDrDirID = desktopDirID;
			cPB.dirInfo.ioNamePtr = NULL;
			cPB.dirInfo.ioFDirIndex = -1;	/* get info on dirID */
			cPB.dirInfo.ioACUser = 0;		/* Note: ioACUser is filler2 in old headers */
			retCode = PBGetCatInfoSync(&cPB);
		}

		/* Should choose desktop button be active? */
		if ( (retCode == noErr) &&
			 (hasDesktopFolder) &&
			 (myDataPtr->SFRPtr->sfFile.parID == 1) )
		{
			/* desktop level is selected and volume has desktop folder */
			
			if ( !myDataPtr->selectOnlyChangable ||
			     ((hPB.volumeParam.ioVAtrb & vcbAtrbLockMask) == 0) &&
			      ((cPB.dirInfo.ioACUser & acUserMakeChanges) == 0) )	/* Note: ioACUser is filler2 in old headers */
			{
				/* it is selectable */
				ActivateButton(theDialog, kSelectDesktopItem);
			}
			else
			{
				/* selectOnlyChangable was true and this location isn't changable */
				InactivateButton(theDialog, kSelectDesktopItem);
			}
		}
		else
		{
			InactivateButton(theDialog, kSelectDesktopItem);
		}
	}
	return retCode;
}

/*****************************************************************************/

/*
**	LastCallCleanup cleans up the data in the StandardFileReply record if
**	SFRPtr->sfGood is true.
*/

static	OSErr	LastCallCleanup(SFDataPtr myDataPtr)
{
	OSErr		retCode;
	CInfoPBRec	cPB;
	Boolean		folderFlag, wasAliasedFlag;
	
	retCode = noErr;
	if ( myDataPtr->SFRPtr->sfGood )
	{
		/* cancel not pressed and no fatal error occured */
		
		/* [<2>] Weirdly, ResolveAliasFile behaves differently depending on whether or
			not File Sharing is turned on.  Specifically, if oldChoice comes in with an
			empty name (from the user clicking the "Select <x>" button with nothing
			selected), ResolveAliasFile will succeed if File Sharing is turned off, but
			fail with a paramErr (-50) if it's on.  Therefore, we do an extra PBGetCatInfo
			to make sure we've got a complete FSSpec. */
		
		if (myDataPtr->oldChoice.name[0] == 0) {
			cPB.dirInfo.ioNamePtr = myDataPtr->oldChoice.name;
			cPB.dirInfo.ioVRefNum = myDataPtr->oldChoice.vRefNum;
			cPB.dirInfo.ioDrDirID = myDataPtr->oldChoice.parID;
			cPB.dirInfo.ioFDirIndex = -1;	/* get info on dirID and return folderName */
			retCode = PBGetCatInfoSync(&cPB);
			assert(retCode == noErr);
			/* It doesn't really matter if it fails, but we'd like to know for debugging purposes. */
			if (retCode == noErr) {
				// Name implicitly copied out, just get the parent ID.
				myDataPtr->oldChoice.parID = cPB.dirInfo.ioDrParID;
			}
		}
		
		/* we have a valid FSSpec, now let's make sure it's not for an alias file */
		retCode = ResolveAliasFile(&(myDataPtr->oldChoice), true, &folderFlag, &wasAliasedFlag);
		if ( retCode == noErr )
		{
			/* Get the folder's fdFlags and script code (if any) */
			cPB.dirInfo.ioNamePtr = myDataPtr->oldChoice.name;
			cPB.dirInfo.ioVRefNum = myDataPtr->oldChoice.vRefNum;
			cPB.dirInfo.ioDrDirID = myDataPtr->oldChoice.parID;
		//	cPB.dirInfo.ioFDirIndex = 0;
			if ( myDataPtr->oldChoice.name[0] != 0 )
				cPB.dirInfo.ioFDirIndex = 0;	/* get info on named directory */
			else
				cPB.dirInfo.ioFDirIndex = -1;	/* get info on dirID and return folderName */
			retCode = PBGetCatInfoSync(&cPB);
			if ( retCode == noErr )
			{
				myDataPtr->SFRPtr->sfFlags = cPB.dirInfo.ioDrUsrWds.frFlags;
				/* Is there a script code? */
				if ( (cPB.dirInfo.ioDrFndrInfo.frScript & 0x80) != 0 )
				{
					myDataPtr->SFRPtr->sfScript = cPB.dirInfo.ioDrFndrInfo.frScript & 0x7f;
				}
				else
				{
					myDataPtr->SFRPtr->sfScript = smRoman;
				}

				/* Make FSSpec from oldChoice to return */
				retCode = FSMakeFSSpec(myDataPtr->oldChoice.vRefNum, myDataPtr->oldChoice.parID, myDataPtr->oldChoice.name, &(myDataPtr->SFRPtr->sfFile));

				if ( retCode == noErr )
				{
					/* Set sfIsVolume and sfIsFolder properly */
					if ( myDataPtr->SFRPtr->sfFile.parID == fsRtParID )
					{
						myDataPtr->SFRPtr->sfIsVolume = true;
					}
					else
					{
						myDataPtr->SFRPtr->sfIsVolume = false;
					}
					myDataPtr->SFRPtr->sfIsFolder = !myDataPtr->SFRPtr->sfIsVolume;
				}
			}
		}
	}
	return retCode;
}

/*****************************************************************************/

/*
**	MyDlgHook:
**		 sets the folder choose and desktop choose button names.
**		 sets oldChoice so MyDlgHook can tell when the choice has changed.
**		 calls LastCallCleanup to set the fields in the reply record.
**		 Maps the folder choose and desktop choose buttons to sfItemOpenButton.
*/

static	pascal	short	MyDlgHook(short item,
								  DialogPtr theDialog,
								  SFDataPtr myDataPtr)
{
	OSErr			retCode;
	HParamBlockRec	hPB;
	
	/* make certain the proper dialog is showing */
	if ( GetWRefCon(theDialog) == sfMainDialogRefCon )
	{
		if ( (myDataPtr->SFRPtr->sfFile.parID == fsRtParID) &&
			 (myDataPtr->SFRPtr->sfFile.name[0] != 0) )
		{
			/* This keeps the MFS file system code from crashing */
			/* (not that there are many MFS disk left) */
			hPB.volumeParam.ioNamePtr = NULL;
			hPB.volumeParam.ioVRefNum = myDataPtr->SFRPtr->sfFile.vRefNum;
			hPB.volumeParam.ioVolIndex = 0;
			retCode = PBHGetVInfoSync(&hPB);
			if ( retCode != noErr)
			{
				myDataPtr->SFRPtr->sfGood = false;
				return sfItemOpenButton;	/* bail out now! */
			}
			if ( hPB.volumeParam.ioVSigWord == 0xd2d7 )
			{
				/* if volume is MFS volume, then append a colon to the name */
				++(myDataPtr->SFRPtr->sfFile.name[0]);
				myDataPtr->SFRPtr->sfFile.name[myDataPtr->SFRPtr->sfFile.name[0]] = ':';
			}
		}
		switch ( item )
		{
		case sfHookLastCall: 
			retCode = LastCallCleanup(myDataPtr);
			if ( retCode != noErr )
			{
				myDataPtr->SFRPtr->sfGood = false;
			}
			break;
		case kSelectFolderItem: 
			/* Leave with the chosen folder. */
			item = sfItemOpenButton;
			break;
		case kSelectDesktopItem: 
			/* The volume's desktop was chosen. Save it for the sfHookLastCall. */
			myDataPtr->oldChoice.name[0] = 0;
			retCode = FindFolder(myDataPtr->SFRPtr->sfFile.vRefNum, kDesktopFolderType,
								 kCreateFolder, &(myDataPtr->oldChoice.vRefNum), &(myDataPtr->oldChoice.parID));
			if ( retCode != noErr )
			{
				myDataPtr->SFRPtr->sfGood = false;
			}
			/* Leave with the chosen desktop folder. */
			item = sfItemOpenButton;
			break;
		default:
			if ( !SameFSSpec(&(myDataPtr->oldChoice), &(myDataPtr->SFRPtr->sfFile)) )
			{
				/* Change buttons if choice has changed. */
				retCode = SetSelectFolderButton(theDialog, myDataPtr);
				if ( retCode != noErr )
				{
					myDataPtr->SFRPtr->sfGood = false;
					return sfItemOpenButton;	/* bail out now! */
				}
				retCode = SetSelectDesktopButton(theDialog, myDataPtr);
				if ( retCode != noErr )
				{
					myDataPtr->SFRPtr->sfGood = false;
					return sfItemOpenButton;	/* bail out now! */
				}

				/* Save the current choice as the old choice for comparison next time. */
				myDataPtr->oldChoice = myDataPtr->SFRPtr->sfFile;
			}
			break;
		}
	}
	return item;
}

/*****************************************************************************/

pascal	void	StandardGetFolder(ConstStr255Param thePrompt,
									 Boolean showInvisibles,
									 Boolean selectOnlyChangable,
									 StandardFileReply *theSFR)
{
	SFTypeList	mySFTypeList;
	Point		thePoint;
	SFDataPtr	myDataPtr;
	OSErr		retCode;
	long		gestaltSFResponse;
	
	/* default = exit with no data. */
	theSFR->sfGood = false;
	
	/* use Gestalt to check for the CustomGetFile call */
	retCode = Gestalt(gestaltStandardFileAttr, &gestaltSFResponse);
	if ( (retCode == noErr) &&
		 ((gestaltSFResponse & (1 << gestaltStandardFile58) != 0)) )
	{
		myDataPtr = (SFDataPtr)NewPtrClear(sizeof(SFData));
		if ( myDataPtr != NULL )
		{
			SetPt(&thePoint, -1, -1);	/* center dialog */
			
			/* point the user data parameter at the reply record so we can get to it later */
			myDataPtr->SFRPtr = theSFR;
	
			/* Set theSFR to the current volume and directory. */
			/* This will make the buttons draw correctly when the dialog appears. */
			theSFR->sfFile.vRefNum = -LMGetSFSaveDisk();
			theSFR->sfFile.parID = LMGetCurDirStore();
			theSFR->sfFile.name[0] = 0;

			/* let the file filter know what to do with invisible folders */
			myDataPtr->showInvisibles = showInvisibles;
	
			/* let user select changable folders only? */
			myDataPtr->selectOnlyChangable = selectOnlyChangable;
	
			/* The following string resources are retrieved before
			** calling CustomGetFile to make things a bit faster. */
			GetIndString(myDataPtr->beforeFolderStr, kGetFolderID, kBeforeFolderStrNum);
			GetIndString(myDataPtr->afterFolderStr, kGetFolderID, kAfterFolderStrNum);
			GetIndString(myDataPtr->beforeDesktopStr, kGetFolderID, kBeforeDesktopStrNum);
			GetIndString(myDataPtr->afterDesktopStr, kGetFolderID, kAfterDesktopStrNum);
			GetIndString(myDataPtr->desktopStr, kGetFolderID, kDesktopStrNum);
	
			ParamText(thePrompt, "\p", "\p", "\p");
	
			/* throw up the dialog */
			{
				FileFilterYDUPP		fileFilter;
				DlgHookYDUPP		dlgHook;
				
				fileFilter = NewFileFilterYDProc(MyCustomFileFilter);
				dlgHook = NewDlgHookYDProc(MyDlgHook);
				
				if (fileFilter && dlgHook) {
					CustomGetFile(fileFilter, -1, mySFTypeList, theSFR, kGetFolderID,
							thePoint, dlgHook, NULL, NULL, NULL, myDataPtr);
				}
				
				DisposeRoutineDescriptor(fileFilter);
				DisposeRoutineDescriptor(dlgHook);
			}
			
			DisposePtr((Ptr) myDataPtr);
		}
	}
}

/*****************************************************************************/

