/*
==========================================================================================

	File:		mstring.c

	Contains:	Pascal- and C-string manipulation routines

	Copyright:	 1998 by Christopher Nebel, all rights reserved.

	Change History (most recent first):

		 <2>	 8/30/98	Neb		Fixed bogus memory access in StringCopy_cc.
		 <1>	 7/23/98	Neb		Spliced off from osmacsup.c.

=========================================================================================
*/

#include "mstring.h"

#include <stddef.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <Types.h>


/* StringCopy
*/
StringPtr StringCopy_pp(StringPtr dst, ConstStr255Param src, size_t dstSize)
{
	assert(dstSize <= 256);
	assert(dstSize == 0 || dstSize > sizeof(void *));
	
	if (src && dst) {
		unsigned char	i, sl, *d;
	
		if (dstSize == 0 || dstSize > 256)
			dstSize = 256;
		
		sl = StrLength(src) + 1;
		if (sl > dstSize - 1) sl = dstSize - 1;
		
		for (i = sl, d = dst; i > 0; i--)
			*d++ = *src++;
		
		*d = '\0';
		dst[0] = sl - 1;
	}
	return dst;
}

StringPtr StringCopy_pc(StringPtr dst, const char *src, size_t dstSize)
{
	assert(dstSize <= 256);
	assert(dstSize == 0 || dstSize > sizeof(void *));
	
	if (src && dst) {
		unsigned char	i, *d;
		
		if (dstSize == 0 || dstSize > 256)
			dstSize = 256;
		
		for (i = 0, d = dst+1; *src != '\0' && i < dstSize - 1; i++)
			*d++ = *src++;
		
		*d = '\0';
		dst[0] = i;
	}
	return dst;
}

char *StringCopy_cp(char *dst, ConstStr255Param src, size_t dstSize)
{
	assert(dstSize == 0 || dstSize > sizeof(void *));
	
	if (src && dst) {
		char   *d = dst;
		size_t	i;			/* Number of bytes to copy, including terminator. */
		
		i = src[0] + 1;
		if (dstSize > 0 && i > dstSize)
			i = dstSize;
		
		src += 1;			/* Skip length byte. */
		for (; i > 1; i--)	/* Copy string contents. */
			*d++ = *src++;
		
		*d = '\0';			/* Terminate destination. */
	}
	return dst;
}

char *StringCopy_cc(char *dst, const char *src, size_t dstSize)
{
	char   *d = dst;
	
	assert(dstSize == 0 || dstSize > sizeof(void *));
	
	if (src && dst) {
		for (d = dst; *src && dstSize > 1; --dstSize)
			*d++ = *src++;
		
		*d = '\0';
	}
	return dst;
}

/* StringCat
*/
StringPtr StringCat_pp(StringPtr dst, ConstStr255Param src, size_t dstSize)
{
	assert(dstSize <= 256);
	assert(dstSize == 0 || dstSize > sizeof(void *));
	
	if (src && dst) {
		unsigned char	*d = &dst[dst[0]+1], sl = StrLength(src), i;
		
		dstSize -= StrLength(dst);
		assert(StrLength(src) < dstSize);
		src += 1;
		
		/* Copy string, leaving room for null terminator. */
		for (i = 0; sl > 0 && dstSize > 1; --sl, --dstSize)
			*d++ = *src++, i++;
		
		*d = '\0';
		dst[0] += i;
	}	
	return dst;
}

StringPtr StringCat_pc(StringPtr dst, const char *src, size_t dstSize)
{
	assert(dstSize <= 256);
	assert(dstSize == 0 || dstSize > sizeof(void *));
	
	if (src && dst) {
		unsigned char	*d = &dst[dst[0]+1], sl;
		
		dstSize -= StrLength(dst);
		assert(strlen(src) < dstSize);
		
		for (sl = 0; *src && dstSize > 1; ++sl, --dstSize)
			*d++ = *src++;
		
		*d = '\0';
		dst[0] += sl;
	}
	return dst;
}

char *StringCat_cp(char *dst, ConstStr255Param src, size_t dstSize)
{
	assert(dstSize == 0 || dstSize > sizeof(void *));
	
	if (src && dst) {
		size_t	dl = strlen(dst), n = StrLength(src);
		char   *d;
		
		dstSize -= dl;
		assert(StrLength(src) < dstSize);
		src += 1;
		
		for (d = dst + dl; n > 0 && dstSize > 1; --n, --dstSize)
			*d++ = *src++;
		
		*d = '\0';
	}
	return dst;
}

char *StringCat_cc(char *dst, const char *src, size_t dstSize)
{
	assert(dstSize == 0 || dstSize > sizeof(void *));
	
	if (src && dst) {
		size_t	dl = strlen(dst);
		char   *d;
		
		dstSize -= strlen(dst);
		assert(strlen(src) < dstSize);
		
		for (d = dst + dl; *src && dstSize > 1; --dstSize)
			*d++ = *src++;
		
		*d = '\0';
	}
	
	return dst;
}

/* StringCompare
*/
int StringCompare_pp(ConstStr255Param a, ConstStr255Param b)
{
	char	ca[256], cb[256];
	
	StringCopy_cp(ca, a, sizeof(ca));
	StringCopy_cp(cb, b, sizeof(cb));
	return strcmp(ca, cb);
}

int StringCompare_pc(ConstStr255Param a, const char *b)
{
	char	ca[256];
	
	StringCopy_cp(ca, a, sizeof(ca));
	return strcmp(ca, b);
}

int StringCompare_cp(const char *a, ConstStr255Param b)
{
	char	cb[256];
	
	StringCopy_cp(cb, b, sizeof(cb));
	return strcmp(a, cb);
}

/* StringEqual
*/
bool StringEqual_pp(ConstStr255Param a, ConstStr255Param b)
{
	char	ca[256], cb[256];
	
	StringCopy_cp(ca, a, sizeof(ca));
	StringCopy_cp(cb, b, sizeof(cb));
	return strcmp(ca, cb) == 0;
}

bool StringEqual_pc(ConstStr255Param a, const char *b)
{
	char	ca[256];
	
	StringCopy_cp(ca, a, sizeof(ca));
	return strcmp(ca, b) == 0;
}

bool StringEqual_cp(const char *a, ConstStr255Param b)
{
	return StringEqual_pc(b, a);
}

bool StringEqual_cc(const char *a, const char *b)
{
	return strcmp(a, b) == 0;
}

/* psprintf
*/
int psprintf(StringPtr buf, const char *fmt, ...)
{
	va_list		arg;
	int			len;
	
	va_start(arg, fmt);
	len = vsprintf((char *) &buf[1], fmt, arg);
	va_end(arg);
	
	assert(len < 256);
	buf[0] = (unsigned char) len;
	return len;
}

int pvsprintf(StringPtr buf, const char *fmt, va_list arg)
{
	int			len;
	
	len = vsprintf((char *) &buf[1], fmt, arg);
	
	assert(len < 256);
	buf[0] = (unsigned char) len;
	return len;
}
